﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.Collections;
using System.Drawing;
using MUD.Client.Corba;

namespace MUD.Client
{
    public enum ObjectType
    {
        Player,Monster,Teleporter
    }

    /// <summary>
    /// A simple AI class based on Hillclimbing using manhattan distance
    /// </summary>
    public class AI
    {
        /// <summary>
        /// My own security tooken
        /// </summary>
        private int token;
        /// <summary>
        /// My self
        /// </summary>
        private GameObject myself = null;

        /// <summary>
        /// List of players in the current room
        /// </summary>
        private C5.TreeDictionary<int, GameObject> players = new C5.TreeDictionary<int, GameObject>();
        /// <summary>
        /// List of transporters in the current room
        /// </summary>
        private C5.TreeDictionary<int, GameObject> transporter = new C5.TreeDictionary<int, GameObject>();

        /// <summary>
        /// Server connection
        /// </summary>
        private ServerConnection server = null;
        
        /// <summary>
        /// System timer to do an action on
        /// </summary>
        private System.Timers.Timer timer = null;

        /// <summary>
        /// Some booleans
        /// -Whether to include transporters in the calculation
        /// -Check whether the Ai is running
        /// -Check to see whether the Ai is initialized
        /// -Checks to see if we're stuck behind and obstacle and if so, go to an alternate location
        /// -Check to make sure we don't use a transporter 2x after eachother. So we don't end up in and endless
        /// teleportation loop
        /// </summary>
        private bool includeTransporter,running,ready,needsRshift,teleported;

        private int last_x, last_y, last_direction,roomWidth,roomHeight;

        /// <summary>
        /// Create a new Ai Instance, (Supports multiple Ai per client)
        /// </summary>
        /// <param name="interval">Interval on which an action should be taken</param>
        /// <param name="server">The server connection to use</param>
        public AI(double interval, ServerConnection server)
        {
            this.server = server;
            this.timer = new System.Timers.Timer(interval);
            this.timer.Elapsed += new System.Timers.ElapsedEventHandler(timer_Elapsed);
        }

        /// <summary>
        /// Timer elapsed, Time to do an action is ready
        /// </summary>
        /// <param name="sender">Object which invoked the call</param>
        /// <param name="e">Extra parameter</param>
        void timer_Elapsed(object sender, System.Timers.ElapsedEventArgs e)
        {
            if (!ready) return;

            int distTarget, locTarget;
            bool isTransporter;

            Point newpoint = calculateNextMove(includeTransporter,out distTarget,out locTarget,out isTransporter);

            //check to see if we're 1 square away from the target and facing him
            if (distTarget == 1 && locTarget==last_direction && !isTransporter)
            {
                server.Attack();
            }
            else
            {
                server.Move(newpoint.X, newpoint.Y);
            }
        }

        /// <summary>
        /// Start the Ai
        /// </summary>
        /// <param name="includeTransporter">Whether the Ai is able to use a transporter</param>
        public void StartAI(bool includeTransporter)
        {
            if (running) return;

            running = true;
            this.includeTransporter = includeTransporter;
            timer.Start();
        }

        /// <summary>
        /// Stop the Ai
        /// </summary>
        public void StopAI()
        {
            if (!running) return;

            running = false;
            timer.Stop();
        }

        /// <summary>
        /// Initialize the Ai using a room Init event
        /// </summary>
        /// <param name="xml">Room init event XML</param>
        /// <param name="includeMonsters">whether we should chase after room monsters as well</param>
        internal void init(string xml,bool includeMonsters)
        {
            XmlDocument doc = new XmlDocument();
            doc.LoadXml(xml);

            teleported = ready;

            this.token = server.getSessionId();

            roomHeight=int.Parse(doc.SelectSingleNode("//tiles/@height").InnerText);
            roomWidth = int.Parse(doc.SelectSingleNode("//tiles/@width").InnerText);

            myself=new GameObject(doc.SelectSingleNode("//player[@id=" + token +"]"), ObjectType.Player);
            XmlNodeList nodes=doc.SelectNodes("//player[@id!=" + token +"]" + (includeMonsters ? " | //monster" : ""));
            players.Clear();
            transporter.Clear();
            foreach (XmlNode item in nodes)
            {
                players.Add(int.Parse(item.Attributes["id"].InnerText), new GameObject(item,ObjectType.Player));
            }

            nodes = doc.SelectNodes("//transporter");
            foreach (XmlNode item in nodes)
            {
                transporter.Add(int.Parse(item.Attributes["id"].InnerText), new GameObject(item,ObjectType.Teleporter));
            }

            if (includeTransporter && teleported)
            {
                int target;
                GameObject teleporter = calculateClosestFeat(true, out target);
                if (teleporter.Type == ObjectType.Teleporter){
                    teleporter.validMove = false;
                    transporter[teleporter.Id] = teleporter;
                }
            }

            ready = true;

        }

        /// <summary>
        /// Calculate the closes GameObject to me
        /// </summary>
        /// <param name="includeTransporter">Whether to include transporters in the calculation</param>
        /// <param name="target">Reference to an integer which will contain the distance to the target in the end</param>
        /// <returns>Closes GameObject</returns>
        internal GameObject calculateClosestFeat(bool includeTransporter, out int target)
        {
            GameObject[] locs = players.Values.ToArray();
            if (includeTransporter)
            {
                GameObject[] total = new GameObject[locs.Length + transporter.Count];
                Array.Copy(locs, total, locs.Length);
                Array.Copy(transporter.Values.ToArray(), 0, total, locs.Length, transporter.Count);
                locs = total;
            }

            int min = int.MaxValue;
            GameObject mDist = null;

            foreach (GameObject item in locs)
            {
                if(!item.validMove) continue;

                int dist = Math.Abs(item.X - myself.X) + Math.Abs(item.Y - myself.Y);
                if (dist < min)
                {
                    min = dist;
                    mDist = item;
                }
            }

            target = min;
            locs = null;
            return mDist;
        }

        /// <summary>
        /// Calculate the next move the player should make
        /// </summary>
        /// <param name="includeTransporter">Whether to include transports</param>
        /// <param name="target">Reference to an integer which will contain the distance to the target in the end</param>
        /// <param name="direction">Reference to an integer that contains the direction the move is going to take</param>
        /// <param name="isTransporter">Reference to a boolean which indicates if the object choosen was a teleporter</param>
        /// <returns>new X and Y coordinate to move to.</returns>
        internal Point calculateNextMove(bool includeTransporter, out int target,out int direction,out bool isTransporter)
        {
            GameObject who = calculateClosestFeat(includeTransporter,out target);

            isTransporter = who.Type == ObjectType.Teleporter;

            if (who == null)
            {
                ready = false;
                direction = -1; //target point is invalid, we have no one to go after
                return new Point(myself.X, myself.Y);
            }

            Point vec = new Point(who.X-myself.X,who.Y-myself.Y);

            if (Math.Abs(vec.X) > Math.Abs(vec.Y))
            {
                //Move Horizontally
                int dx = Math.Sign(vec.X); /// (Math.Abs(vec.X));

                direction = dx > 0 ? 2 : 3; 

                if (needsRshift)
                {
                    needsRshift = false;
                    int dy = (myself.Y == roomHeight ? -1 : 1);
                    return new Point(myself.X, myself.Y + dy);
                }
                return new Point(myself.X + dx, myself.Y);
            }
            else
            {
                //Move Vertically
                int dy = Math.Sign(vec.Y); /// (Math.Abs(vec.Y));
                
                direction = dy > 0 ? 0 : 1;

                if (needsRshift)
                {
                    needsRshift = false;
                    int dx = (myself.X == roomWidth ? -1 : 1);
                    return new Point(myself.X+dx, myself.Y);
                }
                return new Point(myself.X, myself.Y+dy);
            }
        }

        /// <summary>
        /// Set my security Token
        /// </summary>
        /// <param name="p">security token</param>
        internal void setId(int p)
        {
            this.token = p;
        }

        /// <summary>
        /// Move myself to a new location on the board. And checks if we're doing the same move over and over again
        /// </summary>
        /// <param name="id">my own id</param>
        /// <param name="x">new x</param>
        /// <param name="y">new y</param>
        /// <param name="direction">new direction</param>
        internal void move(int id, int x, int y, int direction)
        {
            if (id == myself.Id)
            {
                myself.Direction = direction;
                myself.X = x;
                myself.Y=y;

                if (last_direction == direction && last_x == x && last_y == y)
                {
                    needsRshift = true;
                }
                else
                {
                    last_direction = direction;
                    last_x = x;
                    last_y = y;
                }
                return;
            }

            if (!players.Contains(id)) return;

            GameObject updated = players[id];
            updated.Direction = direction;
            updated.X = x;
            updated.Y = y;
            players[id] = updated;
        }

        /// <summary>
        /// Removes a GameObject from the board
        /// </summary>
        /// <param name="p">GameObject id to be removed</param>
        internal void remove(int p)
        {
            if (p == myself.Id)
            {
                ready = false;
                myself = null;
            }
            else
            {
                if (players.Contains(p))
                    players.Remove(p);

                teleported = false;
            }
        }

        /// <summary>
        /// Add a new game object to the board.
        /// </summary>
        /// <param name="p">XML of the new object</param>
        internal void add(string p)
        {
            XmlDocument doc=new XmlDocument();
            doc.LoadXml(p);

            XmlNode node=doc.SelectSingleNode("//player");

            int id=int.Parse(node.Attributes["id"].InnerText);

            if (id == token)
            {
                myself = new GameObject(node,ObjectType.Player);
            }
            else
            {
                players.Add(id, new GameObject(node,ObjectType.Player));
            }

            if(myself!=null)
                ready = true;
        }
    }

    /// <summary>
    /// Holder class for GameObjects
    /// <remarks>
    /// Contains the x, y and direction and whether the object is a valid object to choose
    /// </remarks>
    /// </summary>
    public class GameObject
    {
        private int x, y, direction,id;
        private bool validMoves=true;
        private ObjectType type;

        public GameObject(XmlNode info,ObjectType type)
        {
            direction = int.Parse(info.SelectSingleNode("direction").InnerText);
            x = int.Parse(info.SelectSingleNode("x").InnerText);
            y = int.Parse(info.SelectSingleNode("y").InnerText);
            id = int.Parse(info.SelectSingleNode("@id").InnerText);
            this.type=type;
        }

        public int Id { get { return id; } }
        public int Direction { 
            get { return direction; }
            set { direction = value; }
        }
        public int X
        {
            get { return x; }
            set { x = value; }
        }
        public int Y
        {
            get { return y; }
            set { y = value; }
        }

        public bool validMove
        {
            get { return validMoves; }
            set { validMoves = value; }
        }

        public ObjectType Type
        {
            get { return type; }
        }
    }
}
