﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using MUD.Client.Exceptions;

namespace MUD.Client.Corba
{
    public partial class ServerConnection
    {
        /// <summary>
        /// The security token that's been given by the server
        /// </summary>
        private int security_token = 0;

        /// <summary>
        /// Checks if the user is logged in to the system, if not throw an exception
        /// </summary>
        /// <exception cref="AuthenticationException">User is not logged in</exception>
        private void CheckAuthentication()
        {
            if (security_token == 0)
                throw new AuthenticationException("You are not logged in to the game");
        }

        /// <summary>
        /// Checks to see if the user is logged in
        /// </summary>
        /// <returns>Whether the user is logged in or not</returns>
        internal bool isLoggedIn()
        {
            return security_token != 0;
        }

        /// <summary>
        /// Try and perform an action on the server
        /// </summary>
        /// <param name="action_index">The id of the action to be performed</param>
        /// <param name="object_id">The id of the action to perform</param>
        /// <returns>Whether the action has succeeded or not</returns>
        public bool Action(int action_index,int object_id)
        {
            CheckAuthentication();
            return server.action(security_token, object_id,action_index);
        }

        /// <summary>
        /// Try to perform an attack on something/someone on the server
        /// </summary>
        /// <returns>Whether the attack has be</returns>
        public bool Attack()
        {
            CheckAuthentication();
            return server.attack(security_token);
        }

        /// <summary>
        /// Try and authenticate to the game server
        /// </summary>
        /// <param name="username">Username for a player</param>
        /// <param name="password">Password for the player</param>
        /// <returns>Whether the login was successful</returns>
        public bool Login(String username, String password)
        {
            int results = server.login(
                    Utils.Convertion.CreateFixedLengthCharArray(username, 64),
                    Utils.Convertion.CreateFixedLengthCharArray(password, 32),
                    client);

            bool loggedin = results != 0;

            if (loggedin)
                security_token = results;

            return loggedin;
        }

        /// <summary>
        /// Logout from the game server
        /// </summary>
        public void Logout()
        {
            CheckAuthentication();
            server.logout(security_token);
            security_token = 0;
        }

        /// <summary>
        /// Try and perform a move operation on your player
        /// </summary>
        /// <param name="x">New X Coordinate</param>
        /// <param name="y">New Y Coordinate</param>
        /// <returns>Whether the move was successful or not</returns>
        public bool Move(int x, int y)
        {
            CheckAuthentication();
            return server.move(security_token, x, y);
        }

        /// <summary>
        /// Try and use an item
        /// </summary>
        /// <param name="item_id">Id of the item to be used</param>
        /// <returns>Whether the operation was successful or not</returns>
        public Boolean UseItem(int item_id)
        {
            CheckAuthentication();
            return server.useItem(security_token, item_id);
        }

        /// <summary>
        /// Gets a list of active users
        /// </summary>
        /// <returns>A List of currently active users</returns>
        public String[] GetUsers()
        {
            CheckAuthentication();
            char[][] result = server.getUsers(security_token);

            String[] realResult = new String[result.Length];

            for (int i = 0; i < result.Length; i++)
            {
                realResult[i] = new String(result[i]).Trim();
            }

            return realResult;
        }

        /// <summary>
        /// Sends a chat message to the server
        /// </summary>
        /// <param name="message">Messafe to be send</param>
        public void SendMessage(String message)
        {
            CheckAuthentication();
            server.sendMessage(security_token,
                    Utils.Convertion.CreateFixedLengthCharArray(message,256));
        }
    }
}
