﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using mud.common.idl.clients;
using System.Threading;

namespace MUD.Client
{
    public class WaitQueue : IDisposable
    {
        /// <summary>
        /// The queue that holds all the events
        /// </summary>
        private Queue<_Event> events;

        /// <summary>
        /// The thread collection
        /// </summary>
        private Thread[] threads;

        /// <summary>
        /// The variable that controls the state of everything
        /// </summary>
        private Boolean running;

        /// <summary>
        /// Pointer to the function to call to set an event message
        /// </summary>
        private System.Threading.ParameterizedThreadStart receiver;

        /// <summary>
        /// Create a new WaitQueue
        /// </summary>
        /// <param name="threadCount">Amount of threads to assign to the thread queue</param>
        public WaitQueue(int threadCount)
        {
            events = new Queue<_Event>();

            threads = new Thread[threadCount];
            running = true;
            for (int i = 0; i < threads.Length; i++)
            {
                threads[i] = new Thread(new ThreadStart(ProcessEvent));
                threads[i].Start();
            }
        }

        /// <summary>
        /// Method where the threads run
        /// </summary>
        internal void ProcessEvent()
        {
            while (running)
            {
                try
                {
                    _Event item;

                    System.Diagnostics.Debug.WriteLine("Thread Entered: " + Thread.CurrentThread.GetHashCode());
                    Monitor.Enter(events);

                    while (events.Count() == 0 && running)
                        Monitor.Wait(events);

                    item = events.Dequeue();

                    receiver.Invoke(item);

                    System.Diagnostics.Debug.WriteLine("Event: " + item.ToString());
                }
                catch (Exception)
                {
                    return;
                }
                finally
                {
                    if(events!=null)
                        Monitor.Exit(events);
                }
            }
        }

        /// <summary>
        /// Stop processing
        /// </summary>
        public void Stop()
        {
            lock (this)
            {
                if (!running) return;

                running = false;

                for (int i = 0; i < threads.Length; i++)
                {
                    threads[i].Interrupt();
                    threads[i].Abort();
                    threads[i] = null;
                    
                    Monitor.Enter(events);
                    Monitor.PulseAll(events);
                    Monitor.Exit(events);
                }
                events = null;
                threads = null;
            }
        }

        /// <summary>
        /// Add a new event to the queue
        /// </summary>
        /// <param name="evnt">Event to add</param>
        public void Push(_Event evnt){
            if (events == null) return; //we were disposed
                Monitor.Enter(events);

                events.Enqueue(evnt);
                Monitor.Pulse(events);

                Monitor.Exit(events);
        }

        public void Dispose()
        {
            Stop();
        }

        internal void setReceiver(ParameterizedThreadStart eventstart)
        {
            this.receiver = eventstart;
        }
    }
}
