package mud.client;

import static mud.common.Constants.FRAME_TIME;

import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

import mud.client.corba.ServerConnection;
import mud.client.messages.Message;
import mud.client.view.GameView;
import mud.common.events.ServerEventHandler;
import mud.common.idl.clients.Event;
import mud.common.objects.Player;
import mud.common.room.Room;

public class GameLoop implements Runnable, KeyListener {

	//Thread
	private Thread thread;
	private boolean running; //is thread running?
	private boolean quit;    //when true, signals thread should stop ASAP
	
	//State
	private GameView view;	
	private Room room;
	private LocalPlayer localPlayer;
	private int playerId;
	
	//User Input
	private UserInput input;
	private List<Integer> keysPressed;
	private List<Integer> keysReleased;
	
	//Networking
	private List<Event> receiveQueue;
	private List<Message> sendQueue;
	private SendLoop sendLoop;
	
	public GameLoop(GameView view) {
		this.view = view;

		sendLoop = new SendLoop();
	}
	
	//Functions
	public void start(ServerConnection sc) {
		if (running) throw new IllegalStateException("GameLoop.start() :: Already running");
		running = true;
		quit = false;

		view.start();
		
		//Create state
		localPlayer = new LocalPlayer(view);
		
		//Input related
		input = new UserInput();
		keysPressed = new LinkedList<Integer>();
		keysReleased = new LinkedList<Integer>();
		view.addKeyListener(this);

		//Networking
		receiveQueue = new LinkedList<Event>();
		sendQueue = new ArrayList<Message>();
		sendLoop.start(sc);
		
		//Start thread
		thread = new Thread(this, "Client::GameLoop");		
		thread.start();
	}
	public void stop() {			
		quit = true;
	}
	
	public void run() {
		long time;
		
		while (!quit) {
			time = System.currentTimeMillis();
			
			synchronized (input) {
				//Put all new keyevents into the userinput object for this frame
				input.add(keysPressed, keysReleased);
				keysPressed.clear();
				keysReleased.clear();
			}
			
			update();
			if (quit) break;
			
			view.draw(room);
			
			//Limit the frame rate
			long dt = System.currentTimeMillis() - time;
			if (dt < FRAME_TIME) {
				try {
					Thread.sleep(FRAME_TIME - dt);
				} catch (InterruptedException ie) {}
			}
		}
		
		//Stop Networking
		receiveQueue = null;
		sendQueue = null;
		sendLoop.stop();
		
		//Dispose state
		room = null;
		playerId = 0;
		localPlayer = null;
		
		//Dispose input related
		keysPressed = keysReleased = null;
		input = null;
		view.removeKeyListener(this);

		view.stop();
		
		running = false;
	}

	public void addServerEvent(Event event) {
		synchronized (receiveQueue) {
			receiveQueue.add(event);
		}
	}
	
	/** Processes any received server events */
	protected void updateServerEvents() {
		Event event;
		do {
			event = null;
			synchronized (receiveQueue) {
				if (!receiveQueue.isEmpty()) {
					event = receiveQueue.remove(0);
				}
			}
			
			if (event == null) break;

			try {
				room = ServerEventHandler.process(room, event);
				if (room != null) {
					localPlayer.setPlayer((Player)room.getObjectById(playerId));
				}
			} catch (RuntimeException re) {
				re.printStackTrace();
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		} while (event != null);			
	}
	public void update() {
		sendQueue.clear();

		//Process stuff that came from the server
		updateServerEvents();
		
		//Do any local updates here...

		//Process user input
		localPlayer.update(input, sendQueue);
		sendLoop.update(sendQueue);
	}
	
	/** Put key events in a buffer */
	public void keyPressed(KeyEvent e) {
		int key = e.getKeyCode();
		
		if (key == KeyEvent.VK_ESCAPE) {
			//Panic button
			System.exit(1);
		} else {
			synchronized (input) {
				keysReleased.remove((Object)e.getKeyCode());
				keysPressed.add(e.getKeyCode());
			}
		}
	}
	/** Put key events in a buffer */
	public void keyReleased(KeyEvent e) {
		synchronized (input) {
			keysReleased.add(e.getKeyCode());
		}
	}
	/** Put key events in a buffer */
	public void keyTyped(KeyEvent e) {
	}
	
	//Getters
	
	//Setters
	public void setPlayerId(int pid) {
		playerId = pid;		
	}

}
