package mud.client.corba;

import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;

import mud.common.Utils;
import mud.common.exceptions.AuthenticationException;
import mud.common.idl.clients.Client;
import mud.common.idl.clients.ClientHelper;
import mud.common.idl.servers.Server;
import mud.common.idl.servers.ServerHelper;
import mud.common.idl.servers.ServerPackage.InvalidSecurityToken;

import org.omg.CORBA.ORB;
import org.omg.CORBA.ORBPackage.InvalidName;
import org.omg.CosNaming.NamingContextExt;
import org.omg.CosNaming.NamingContextExtHelper;
import org.omg.CosNaming.NamingContextPackage.CannotProceed;
import org.omg.CosNaming.NamingContextPackage.NotFound;
import org.omg.PortableServer.POA;
import org.omg.PortableServer.POAHelper;
import org.omg.PortableServer.POAManagerPackage.AdapterInactive;
import org.omg.PortableServer.POAPackage.ServantNotActive;
import org.omg.PortableServer.POAPackage.WrongPolicy;

/**
 * Create a new ORB and initialize and register the classes
 * @author Tamar Christina
 *
 */
public class ServerConnection {

    private ORB orb = null;
    private String[] args;
    private Server serv;
    private int security_token=0;
    private ClientImpl client;
    private Client clientRef;

    /**
     * Create a new ServerConnection Instance and pass along the command line arguments
     * @param args CommandLine arguments for the program. Used to initialise corba.
     */
    public ServerConnection(String[] args) {
    	this.args = args;
    }

    /**
     * Create and initialize the orb
     */
    private void createAndInitOrb() {
	    try {
		Properties properties = new Properties();
		orb = ORB.init(args, properties);
		org.omg.CORBA.Object objRef = orb.resolve_initial_references("NameService");

		// registering server part
		NamingContextExt ncRef = NamingContextExtHelper.narrow(objRef);
		String bname = "GameServer";
		serv = ServerHelper.narrow(ncRef.resolve_str(bname));
		
		// registering client part
		POA rootPoa = POAHelper.narrow(orb
				.resolve_initial_references("RootPOA"));
		rootPoa.the_POAManager().activate();

		client = new ClientImpl();
		org.omg.CORBA.Object ref = rootPoa.servant_to_reference(client);
		clientRef = ClientHelper.narrow(ref);
		
	    } catch (NotFound ex) {
		Logger.getLogger(ServerConnection.class.getName()).log(Level.SEVERE, null, ex);
	    } catch (CannotProceed ex) {
		Logger.getLogger(ServerConnection.class.getName()).log(Level.SEVERE, null, ex);
	    } catch (org.omg.CosNaming.NamingContextPackage.InvalidName ex) {
		Logger.getLogger(ServerConnection.class.getName()).log(Level.SEVERE, null, ex);
	    } catch (InvalidName ex) {
		Logger.getLogger(ServerConnection.class.getName()).log(Level.SEVERE, null, ex);
	    } catch (AdapterInactive e) {
	    Logger.getLogger(ServerConnection.class.getName()).log(Level.SEVERE, null, e);
		} catch (ServantNotActive e) {
		Logger.getLogger(ServerConnection.class.getName()).log(Level.SEVERE, null, e);
		} catch (WrongPolicy e) {
		Logger.getLogger(ServerConnection.class.getName()).log(Level.SEVERE, null, e);
		}
    }
    
    /**
     * Get a reference to the current ClientImp class
     * @return ClientImpl class, or Null if connection has not yet been made.
     */
    public ClientImpl getClient(){
    	return client;
    }

    public int getSecurityToken() {
    	return security_token;
    }
    
    /**
     * Stop the orb and kill the main loop
     */
    public void stop() {
	if (orb != null) {
	    orb.shutdown(false);
	}
    }

    /**
     * Start the main loop thread
     */
    public void start() {
	createAndInitOrb();
    }
    
    /**
     * Checks if the user is logged in to the system, if not throw an exception
     * @throws mud.common.exceptions.AuthenticationException User is not logged in
     */
    private void checkAuthenticated() throws AuthenticationException {
	if(security_token==0)
	    throw new AuthenticationException("You are not logged in to the game");
    }

    /**
     * Try and perfom an action on the server
     * @param action_index The id of the action to be performed
     * @return Whether the action has succeeded or not
     * @throws AuthenticationException if user is not logged in
     * @throws InvalidSecurityToken 
     * @throws InvalidSecurityToken When this method is called with an invalid token
     * @throws InvalidSecurityToken 
     */
    public boolean action(int object_id, int action_index) throws AuthenticationException, InvalidSecurityToken {
	checkAuthenticated();
	return serv.action(security_token, object_id, action_index);
    }

    /**
     * Try to perform an attack on something/someone on the server
     * @return Whether the attack has succeeded or not
     * @throws AuthenticationException if user is not logged in
     * @throws InvalidSecurityToken When this method is called with an invalid token
     */
    public boolean attack() throws AuthenticationException, InvalidSecurityToken {
	checkAuthenticated();
	return serv.attack(security_token);
    }

    /**
     * Try and authenticate to the game server
     * @param username Username for a player
     * @param password Password for the player
     * @return Whether the login was successful 
     */
    public boolean login(String username, String password) {
	int result= serv.login(
		Utils.createFixedLengthCharArray(username, 64),
		Utils.createFixedLengthCharArray(password, 32),
		clientRef);
	boolean loggedin = result != 0;
	
	if(loggedin)
	    security_token=result;
	
	return loggedin;
    }

    /**
     * Logout from the game server
     * @throws AuthenticationException if user is not logged in
     * @throws InvalidSecurityToken When this method is called with an invalid token
     */
    public void logout() throws AuthenticationException, InvalidSecurityToken {
	checkAuthenticated();
	serv.logout(security_token);
	security_token=0;
    }

    /**
     * Try and perform a move operation on your player 
     * @param x New X coordinate
     * @param y New Y coordinate
     * @return Whether the move was successful or not
     * @throws AuthenticationException if user is not logged in
     * @throws InvalidSecurityToken When this method is called with an invalid token
     */
    public boolean move(int x, int y) throws AuthenticationException, InvalidSecurityToken {
	checkAuthenticated();
	return serv.move(security_token, x, y);
    }

    /**
     * Try and use an item
     * @param item_id Id of the item to be used
     * @return Whether the operation was successful or not
     * @throws AuthenticationException if user is not logged in
     * @throws InvalidSecurityToken When this method is called with an invalid token
     */
    public boolean useItem(int item_id) throws AuthenticationException, InvalidSecurityToken {
	checkAuthenticated();
	return serv.useItem(security_token, item_id);
    }
    
    /**
     * Gets a list of active users
     * @return A List of currently active users
     * @throws AuthenticationException if user is not logged in
     * @throws InvalidSecurityToken When this method is called with an invalid token
     */
	public String[] getUsers() throws InvalidSecurityToken, AuthenticationException {
		checkAuthenticated();
		char[][] result=serv.getUsers(security_token);

		String[] realResult=new String[result.length];
		
		for (int i = 0; i < result.length; i++) {
			realResult[i]=new String(result[i]).trim();
		}
		
		return realResult;
	}

	/**
	 * Sends a chat message to the server
	 * @param message Message to be send
     * @throws AuthenticationException if user is not logged in
     * @throws InvalidSecurityToken When this method is called with an invalid token
	 */
	public void sendMessage(String message) throws InvalidSecurityToken, AuthenticationException {
		checkAuthenticated();
		serv.sendMessage(security_token, Utils.createFixedLengthCharArray(message, 256));
	}
}
