package mud.common.loader;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.util.Iterator;

import mud.common.image.ImageRect;
import mud.common.items.Armor;
import mud.common.items.Item;
import mud.common.items.UsableItem;
import mud.common.items.Weapon;
import mud.common.objects.Creature;
import mud.common.objects.GameObject;
import mud.common.objects.Inventory;
import mud.common.objects.Monster;
import mud.common.objects.Player;
import mud.common.objects.Transporter;
import mud.common.objects.Treasure;
import mud.common.Direction;
import mud.common.room.GameTile;
import mud.common.room.Room;
import mud.common.xml.XmlElement;
import mud.common.xml.XmlReader;

/** Recreates a room (and other objects) from an xml representation */
public class RoomLoader {

	public RoomLoader() {		
	}
	
	public Room loadRoom(String xmlString) throws IOException {
		ByteArrayInputStream in = new ByteArrayInputStream(xmlString.getBytes());
		
		Room room = null;
		try {
			XmlReader reader = new XmlReader();
			room = parseRoom(reader.read(in).getFirstElementChild());
		} catch (Throwable t) {
			throw new IOException(t);
		} finally {
			in.close();
		}
		
		return room;
	}
	public GameObject loadObject(String xmlString) throws IOException {
		ByteArrayInputStream in = new ByteArrayInputStream(xmlString.getBytes());
		
		GameObject object = null;
		try {
			XmlReader reader = new XmlReader();
			object = parseGameObject(reader.read(in).getFirstElementChild());
		} catch (Throwable t) {
			throw new IOException(t);
		} finally {
			in.close();
		}
		
		return object;
	}
	public Item loadItem(String xmlString) throws IOException {
		ByteArrayInputStream in = new ByteArrayInputStream(xmlString.getBytes());
		
		Item item = null;
		try {
			XmlReader reader = new XmlReader();
			item = parseItem(reader.read(in).getFirstElementChild());
		} catch (Throwable t) {
			throw new IOException(t);
		} finally {
			in.close();
		}
		
		return item;
	}
	
	protected Room parseRoom(XmlElement roomE) {
		XmlElement tilesE = roomE.getChild("tiles");
		int w = getIntAttr(tilesE, "width");
		int h = getIntAttr(tilesE, "height");
		
		GameTile tiles[][] = new GameTile[w][h];
		Iterator<XmlElement> tileElements = tilesE.elementChildren().iterator();
		for (int y = 0; y < h; y++) {
			for (int x = 0; x < w; x++) {
				tiles[x][y] = parseGameTile(tileElements.next()); 
			}
		}
		
		Room room = new Room(getIntAttr(roomE, "id"), tiles);
		
		//Set objects' rooms
		for (int y = 0; y < h; y++) {
			for (int x = 0; x < w; x++) {
				GameObject objects[] = tiles[x][y].getObjects().toArray(new GameObject[0]);
				for (GameObject o : objects) {
					o.setRoom(room);
				}
			}
		}
		return room;
	}
	
	protected GameTile parseGameTile(XmlElement tileE) {		
		GameTile tile = new GameTile(parseImageRect(tileE.getChild("image")));
		tile.setSolid(getBoolean(tileE, "solid"));
		
		for (XmlElement e : tileE.getChild("occupants").elementChildren()) {
			tile.addOccupant(parseGameObject(e));
		}
		
		return tile;
	}
	protected GameObject parseGameObject(XmlElement objectE) {
		String type = objectE.getTagName();
		int id = getIntAttr(objectE, "id");
		ImageRect image = parseImageRect(objectE.getChild("image"));
		
		GameObject object = null;
		if (type.equals("player")) {
			object = new Player(id,objectE.getChild("name").getTextContent(), image, getInt(objectE, "hp"));
		} else if (type.equals("monster")) {
			object = new Monster(id, image, getInt(objectE, "hp"),
					getInt(objectE, "attack"), getInt(objectE, "defense"));
		} else if (type.equals("transporter")) {
			object = new Transporter(id, image, getInt(objectE, "targetRoom"),
					getInt(objectE, "targetX"), getInt(objectE, "targetY"));
		} else if (type.equals("treasure")) {
			object = new Treasure(id, image);
		} else {
			object = new GameObject(id, image);
		}
		
		object.setDirection(parseDirection(objectE.getChild("direction")));
		object.setPos(getInt(objectE, "x"), getInt(objectE, "y"));
				
		if (object instanceof Treasure) {
			Treasure t = (Treasure)object;
			t.getInventory().addAll(parseInventory(objectE.getChild("inventory")));		
		}
		if (object instanceof Creature) {
			Creature c = (Creature)object;
			c.setHp(getInt(objectE, "hp"));
			c.getInventory().addAll(parseInventory(objectE.getChild("inventory")));			
		}
		
		return object;
	}
	
	protected Inventory parseInventory(XmlElement inventoryE) {
		Inventory inventory = new Inventory();
		for (XmlElement e : inventoryE.getChild("items").elementChildren()) {
			inventory.addItem(parseItem(e));
		}
		return inventory;
	}
	
	protected Item parseItem(XmlElement itemE) {
		String type = itemE.getTagName();
		int id = getIntAttr(itemE, "id");

		Item item = null;
		if (type.equals("usableItem")) {
			item = new UsableItem(id, itemE.getAttribute("name"));
		} else if (type.equals("weapon")) {
			item = new Weapon(id, itemE.getAttribute("name"), getInt(itemE, "attack"));
		} else if (type.equals("armor")) {
			item = new Armor(id, itemE.getAttribute("name"), getInt(itemE, "defense"));
		}
		
		return item;
	}
	
	//Utility Functions
	protected Direction parseDirection(XmlElement e) {
		return Direction.fromInt(Integer.parseInt(e.getTextContent()));
	}
	protected ImageRect parseImageRect(XmlElement e) {
		return ImageRect.fromString(e.getTextContent());
	}

	protected boolean getBoolean(XmlElement element, String childName) {
		return Boolean.parseBoolean(element.getChild(childName).getTextContent());
	}
	protected int getInt(XmlElement element, String childName) {
		return Integer.parseInt(element.getChild(childName).getTextContent());
	}
	protected int getIntAttr(XmlElement e, String attributeName) {
		String value = e.getAttribute(attributeName);
		if (value != null) {
			try {
				return Integer.parseInt(value);
			} catch (NumberFormatException nfe) { }
		}
		throw new IllegalArgumentException("Illegal attribute value, expected int: " + value);
	}
}
