package mud.common.room;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import mud.common.objects.GameObject;
import mud.common.xml.XmlElement;

/**
 * Implements a single game room.
 * 
 * @author Pieter van Ede
 */
public class Room extends IdObject {
	
	private GameTile[][] tiles;
	private Set<GameObject> objects;

	/**
	 * Makes a new game room with the specified tables and objects.
	 * 
	 * @param tiles
	 *            all the tiles in the room, where the first dimension is x and
	 *            the second is y.
	 * @param objects
	 *            a list of all game objects in the room.
	 */
	public Room(int id, GameTile[][] tiles) {
		super(id);
		this.tiles = tiles;
				
		objects = new HashSet<GameObject>();
		for (int a = 0; a < tiles.length; a++) {
			for (int b = 0; b < tiles[a].length; b++) {
				for (GameObject go : tiles[a][b].getObjects()) {
					objects.add(go);
				}
			}
		}
	}

	//Functions
	public void addObject(GameObject o) {
		objects.add(o);
		o.setRoom(this);
	}
	public void removeObject(GameObject o) {		
		objects.remove(o);
		o.setRoom(null);
	}
	
	//Getters
	public int getWidth() {
		return tiles.length;		
	}
	public int getHeight() {
		return (getWidth() == 0 ? 0 : tiles[0].length);		
	}
	public Set<GameObject> getObjects() {
		return Collections.unmodifiableSet(objects);
	}
	public GameObject getObjectById(int id) {
		for (GameObject go : objects) {
			if (go.getId() == id) {
				return go;
			}
		}
		return null;
	}
	public GameTile getTile(int x, int y) {
		if (x < 0 || y < 0 || x >= getWidth() || y >= getHeight()) {
			return null;
		}
		return tiles[x][y];
	}
	
	//Setters
	
	//Save Support
	public XmlElement toXml() {
		XmlElement roomE = new XmlElement("room");
		roomE.addAttribute("id", getId());
		
		XmlElement tilesE = roomE.addChild("tiles");
		tilesE.addAttribute("width", getWidth());
		tilesE.addAttribute("height", getHeight());
		for (int y = 0; y < getHeight(); y++) {
			for (int x = 0; x < getWidth(); x++) {
				tilesE.addChild(tiles[x][y].toXml());
			}
		}
		
		return roomE;
	}
}
