package mud.common.xml;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

public class XmlElement extends AbstractXmlNode {

	private String tag;
	
	private Hashtable<String, String> attributes;
	private List<XmlNode> children;
	
	public XmlElement(String tag) {
		this(tag, null);
	}
	public XmlElement(String tag, String textContents) {
		this.attributes = new Hashtable<String, String>();
		this.children = new ArrayList<XmlNode>();
		
		this.tag = XmlUtil.tagEscape(tag);
		if (textContents != null) {
			addChild(new XmlTextNode(textContents));
		}
	}
	
	//Functions
	public void toXmlString(String linePrefix, StringBuilder sb) {
		sb.append(linePrefix);
		sb.append("<");
		sb.append(tag);
		for (Entry<String, String> entry : attributes.entrySet()) {
			sb.append(' ');
			sb.append(entry.getKey());
			sb.append("=\"");
			sb.append(XmlUtil.attributeValueEscape(entry.getValue()));
			sb.append("\"");
		}

		if (children.size() == 0) {
			sb.append("/>");
		} else {
			sb.append(">");
			if (!(children.get(0) instanceof XmlTextNode)) {
				sb.append("\n");
			}
		}
		
		String childLinePrefix = linePrefix + "  ";
		for (int n = 0; n < children.size(); n++) {
			XmlNode prevChild = (n > 0 ? children.get(n - 1) : null);
			XmlNode child = children.get(n);
			XmlNode nextChild = (n < children.size() - 1 ? children.get(n + 1) : null);

			if (prevChild instanceof XmlTextNode) {
				child.toXmlString("", sb);
			} else {
				child.toXmlString(childLinePrefix, sb);
			}
			
			if (!(child instanceof XmlTextNode) && !(nextChild instanceof XmlTextNode)) {
				sb.append("\n");
			}
		}
		
		if (children.size() > 0) {
			if (!(children.get(children.size() - 1) instanceof XmlTextNode)) {
				sb.append(linePrefix);
			}
			sb.append("</");
			sb.append(tag);
			sb.append(">");
		}
	}

	public XmlElement addChild(String tagName) {
		return addChild(tagName, null);
	}
	public XmlElement addChild(String tagName, boolean textContents) {
		return addChild(tagName, Boolean.toString(textContents));
	}
	public XmlElement addChild(String tagName, long textContents) {
		return addChild(tagName, Long.toString(textContents));
	}
	public XmlElement addChild(String tagName, double textContents) {
		return addChild(tagName, Double.toString(textContents));
	}
	public XmlElement addChild(String tagName, String textContents) {
		return (XmlElement)addChild(new XmlElement(tagName, textContents));
	}
	public XmlElement addChild(XmlElement element) {
		return (XmlElement)addChild((XmlNode)element);
	}
	public XmlNode addChild(XmlNode node) {
		return addChild(children.size(), node);
	}

	public XmlElement addChild(int index, String tagName) {
		return addChild(index, tagName, null);
	}
	public XmlElement addChild(int index, String tagName, long textContents) {
		return addChild(index, tagName, Long.toString(textContents));
	}
	public XmlElement addChild(int index, String tagName, String textContents) {
		return (XmlElement)addChild(index, new XmlElement(tagName, textContents));
	}
	public XmlElement addChild(int index, XmlElement element) {
		return (XmlElement)addChild(index, (XmlNode)element);
	}
	public XmlNode addChild(int index, XmlNode node) {
		index = Math.max(0, index);
		
		if (index >= children.size()) {
			children.add(node);
		} else {
			children.add(index, node);
		}
		
		return node;
	}
	
	public void addAttribute(String name, long value) {
		addAttribute(name, Long.toString(value));
	}
	public void addAttribute(String name, String value) {
		attributes.put(XmlUtil.attributeNameEscape(name), value);
	}
	public void removeAttribute(String name) {
		attributes.remove(XmlUtil.attributeNameEscape(name));
	}
	
	//Getters
	public String getTagName() {
		return tag;
	}
	
	public List<XmlNode> getChildren() {
		return Collections.unmodifiableList(children);
	}
	public XmlElement getFirstElementChild() {
		for (XmlNode node : children) {
			if (node instanceof XmlElement) {
				return (XmlElement)node;
			}
		}
		return null;
	}
	public XmlElement getChild(String tagName) {
		for (XmlNode node : children) {
			if (node instanceof XmlElement) {
				XmlElement e = (XmlElement)node;
				if (e.getTagName().equals(tagName)) {
					return e;
				}
			}
		}
		return null;
	}
	public List<XmlElement> getChildren(String tagName) {
		List<XmlElement> list = new ArrayList<XmlElement>();
		for (XmlNode node : children) {
			if (node instanceof XmlElement) {
				XmlElement e = (XmlElement)node;
				if (e.getTagName().equals(tagName)) {
					list.add(e);
				}
			}
		}
		return list;
	}
	public List<XmlElement> getElementChildren() {
		List<XmlElement> list = new ArrayList<XmlElement>();
		for (XmlNode node : children) {
			if (node instanceof XmlElement) {
				list.add((XmlElement)node);
			}
		}
		return list;
	}
	public Iterable<XmlElement> elementChildren() {
		return new Iterable<XmlElement>() {
			public Iterator<XmlElement> iterator() {
				return getElementChildren().iterator();
			}
		};
	}
	
	public Map<String, String> getAttributes() {
		return Collections.unmodifiableMap(attributes);
	}
	public String getAttribute(String name) {
		return attributes.get(name);
	}
	public String getTextContent() {
		StringBuilder sb = new StringBuilder();
		for (XmlNode node : children) {
			sb.append(node.getTextContent());
		}
		return sb.toString();
	}
	
	//Setters
}
