package mud.server;

import java.util.logging.Level;
import java.util.logging.Logger;
import mud.common.idl.clients.Event;
import mud.common.idl.servers.ServerPackage.InvalidSecurityToken;
import mud.server.corba.ClientConnection;
import mud.server.corba.ServerImpl;

/**
 *Event broadcaster singleton class. Use this to send publishing information to
 * the clients
 * 
 * @author Phyx
 */
public class EventBroadcaster {

	private ServerImpl server;
	private int maxThreads;
	private WaitQueue<EventInformation> queue;
	private Thread[] threadpool;
	private boolean running;

	private static EventBroadcaster __instance;

	private EventBroadcaster(ServerImpl server, int maxthreads) {
		this.server = server;
		this.maxThreads = maxthreads;
		this.queue = new WaitQueue<EventInformation>();
	}

	public static EventBroadcaster getInstance(ServerImpl server, int maxthreads)
			throws InstantiationException {

		if (server == null && __instance == null)
			throw new InstantiationException("The Broadcaster has not been initialised.");

		if (__instance == null)
			__instance = new EventBroadcaster(server, maxthreads);

		if (__instance.server == null)
			__instance.server = server;

		return __instance;
	}

	public static EventBroadcaster getInstance() throws InstantiationException {

		return getInstance(null, 0);
	}

	public void sendToRoom(int roomid, Event event) {
		ClientConnection[] clients;
		synchronized (server) {
			clients = server.getConnectionsByRoom(roomid);
		}

		for (int i = 0; i < clients.length; i++) {
			ClientConnection clientConnection = clients[i];
			queue.push(new EventInformation(event, clientConnection));
		}
	}

	public void sendToAll(Event event) {
		ClientConnection[] clients;
		synchronized (server) {
			clients = server.getAllConnections();
		}

		for (int i = 0; i < clients.length; i++) {
			ClientConnection clientConnection = clients[i];
			queue.push(new EventInformation(event, clientConnection));
		}
	}

	public void sendToSingle(Event event, ClientConnection client) {
		queue.push(new EventInformation(event, client));
	}

	private void createThreads(int count) {
		threadpool = new Thread[count];
		running = true;

		Runnable target = new Runnable() {
			public void run() {
				while (running) {
					queue.pop().send(server);
				}
			}
		};

		for (int i = 0; i < threadpool.length; i++) {
			threadpool[i] = new Thread(target);
			threadpool[i].start();
		}
	}

	public void start() {
		if (running)
			return;
		createThreads(maxThreads);
	}

	public void stop() {
		if (!running) {
			return;
		}

		running = false;
		for (int i = 0; i < threadpool.length; i++) {
			threadpool[i] = null;
		}
		threadpool = null;
	}
}

/**
 * Manages an Event
 * 
 * @author Phyx
 */
class EventInformation {
	private Event event;
	private ClientConnection client;

	/**
	 * Creates a new Event Information
	 * 
	 * @param event Event to send
	 * @param client Client to send event to
	 */
	public EventInformation(Event event, ClientConnection client) {
		this.event = event;
		this.client = client;
	}

	/**
	 * Send the message, this method can fail,so do it in a try catch
	 */
	public void send(ServerImpl server) {
		try {
			client.pushEvent(event);
			client = null;
			event = null;
		} catch (Exception e) {
			if (server.isLoggedIn(client.getId())) {
				try {
					server.logout(client.getId());
				} catch (InvalidSecurityToken ex) {
					Logger.getLogger(EventInformation.class.getName()).log(Level.SEVERE, null, ex);
				}
			}
		}
	}
}
