package mud.server.corba;

import java.util.Collection;

import mud.common.Constants;
import mud.common.Direction;
import mud.common.Utils;
import mud.common.idl.clients.Event;
import mud.common.idl.clients.EventPackage.AddObjectEvent;
import mud.common.idl.clients.EventPackage.AttackEvent;
import mud.common.idl.clients.EventPackage.MoveEvent;
import mud.common.idl.clients.EventPackage.UseActionEvent;
import mud.common.idl.clients.EventPackage.UseEvent;
import mud.common.idl.servers.ServerPackage.InvalidSecurityToken;
import mud.common.image.ImageRect;
import mud.common.items.Item;
import mud.common.items.UsableItem;
import mud.common.objects.Creature;
import mud.common.objects.GameObject;
import mud.common.objects.Monster;
import mud.common.objects.Player;
import mud.common.objects.Transporter;
import mud.common.objects.Treasure;
import mud.common.room.GameTile;
import mud.common.room.Room;
import mud.common.room.TileAction;

/** Does all the checks for movement, attacks, etc. */
public class GameLogic {

	private ServerImpl server;

	public GameLogic(ServerImpl server) {
		this.server = server;
	}

	// Functions
	protected void checkLogin(int id) throws InvalidSecurityToken {
		if (!server.isLoggedIn(id)) throw new InvalidSecurityToken();
	}

	public boolean move(int id, int targetX, int targetY) throws InvalidSecurityToken {
		Room room = server.getRoomByUser(id);
		if (room == null) {
			return false; // player was killed while moving
		}
		GameObject object = room.getObjectById(id);
		
		if (!(object instanceof Player)) {
			return false; // Referenced object is not player-controlled
		}
		
		return move((Player)object, targetX, targetY);
	}
	public boolean move(Creature creature, int targetX, int targetY) {
		Room room = creature.getRoom();
		
		if (creature.getHp() <= 0) {
			return false; //Creature is dead
		}
		if (creature.getActivity() != null) {
			return false; //Creature is performing some other action
		}

		int px = creature.getX();
		int py = creature.getY();

		// Continue only if movement is one square in either horizontal OR
		// vertical direction
		if (!((Math.abs(px - targetX) == 1) ^ (Math.abs(py - targetY) == 1))) {
			return false;
		}
		
		//compute direction
		Direction dir;
		if (py == targetY) {
			dir = (px < targetX ? Direction.EAST : Direction.WEST);
		} else {
			dir = (py < targetY ? Direction.SOUTH : Direction.NORTH);
		}

		// Check if the new position is within the room's bounds
		if (targetX < 0 || targetY < 0 || targetX >= room.getWidth() || targetY >= room.getHeight()) {
			return false;
		}

		// Check if the target square is free
		if (!room.getTile(targetX, targetY).isFree()) {
			for (GameObject o : room.getTile(targetX, targetY).getObjects()) {
				if (o instanceof Transporter && creature instanceof Player) {
					Transporter t = (Transporter)o;
					server.setPlayerRoom((Player)creature, t.getTargetRoomId(),
							t.getTargetX(), t.getTargetY(), dir);
					return true;
				}
			}
			// we do not fail completely, we just adjust the players direction
			Event event = new Event();
			event.moveEvent(new MoveEvent(px, py, dir.asInt(), creature.getId()));
			server.executeAndSend(room, event);
			return false;
		}

		Event event = new Event();
		event.moveEvent(new MoveEvent(targetX, targetY, dir.asInt(), creature.getId()));
		server.executeAndSend(room, event);
		return true;
	}

	public boolean attack(int id) throws InvalidSecurityToken {
		Room room = server.getRoomByUser(id);
		if (room == null) {
			return false; // user was previously killed ot removed from room.
		}
		GameObject object = room.getObjectById(id);
		if (!(object instanceof Player)) {
			return false; // Referenced object is not player-controlled
		}
		
		Player player = (Player) object;
		int px = player.getX();
		int py = player.getY();
		
		if (player.getHp() <= 0) {
			return false; //Player is dead
		}
		if (player.getActivity() != null) {
			return false; //Player is performing some other action
		}

		int tx = px + Utils.getDeltaX(player.getDirection(), 1);
		int ty = py + Utils.getDeltaY(player.getDirection(), 1);

		GameTile targetTile = room.getTile(tx, ty);
		if (targetTile == null) {
			// Target tile doesn't exist
			return false;
		}
		Creature target = null;
		for (GameObject o : targetTile.getObjects()) {
			if (o instanceof Creature) {
				target = (Creature) o;
			}
		}
		if (target == null) {
			// No target creature in that location
			return false;
		}
		boolean backAttack = (target instanceof Monster);

		int pATK = player.getAttack();
		int pDEF = player.getDefense();
		int tATK = target.getAttack();
		int tDEF = target.getDefense();
		
		int playerDamage = (backAttack ? Math.max(0, tATK - pDEF) : 0);
		int targetDamage = Math.max(0, pATK - tDEF);
		
		String playerName = server.getUser(player.getId()).getName();
		String targetName = "the monster";
		if (target instanceof Player) {
			targetName = server.getUser(target.getId()).getName();
		}
		
		server.sendMessage(0, "%s attacks %s for %d damage...",
				playerName, targetName, targetDamage);		
		if (backAttack) {
			server.sendMessage(0, "%s attacks %s for %d damage...",
					targetName, playerName, playerDamage);
		}

		Event event = new Event();
		event.attackEvent(new AttackEvent(player.getId(), target.getId(), playerDamage, targetDamage));
		server.executeAndSend(room, event);
		
		if (player.getHp() <= 0) {
			dumpInventory(room, player);
			server.sendMessage(0, "<font color=red>%s has died.</font>", playerName);
		}
		if (target.getHp() <= 0) {
			dumpInventory(room, target);
			server.sendMessage(0, "<font color=red>%s has died.</font>", targetName);
		}
		return true;
	}

	/**
	 * Checks if the creature has items in its inventory and drops them on the
	 * floor.
	 * 
	 * @param room
	 * @param cr
	 */
	private void dumpInventory(Room room, Creature cr) {
		if (!cr.getInventory().isEmpty()) {
			GameTile tile = room.getTile(cr.getX(), cr.getY());
			// TODO: change to treasure sprite
			Treasure lostItems = new Treasure(server.generateId(), new ImageRect("sprite.png", 128, 0,
					Constants.TILE_SIZE, Constants.TILE_SIZE));
			lostItems.setPos(cr.getX(), cr.getY());
			lostItems.getInventory().addAll(cr.getInventory());
			tile.addOccupant(lostItems);
			Event e = new Event();
			e.addObjectEvent(new AddObjectEvent(lostItems.toXml().toXmlString()));
			server.executeAndSend(room, e);
		}
	}
	
	public boolean action(int id, int objectId, int actionId) throws InvalidSecurityToken {
		Room room = server.getRoomByUser(id);
		if (room == null) {
			return false; // player was killed while trying to do an action
		}
		GameObject object = room.getObjectById(id);
		if (!(object instanceof Player)) {
			return false; // Referenced object is not player-controlled
		}
		Player player = (Player) object;
		if (player.getHp() <= 0) {
			return false; //Player is dead
		}
		if (player.getActivity() != null) {
			return false; //Player is performing some other action
		}

		int px = player.getX();
		int py = player.getY();
		int tx = px + Utils.getDeltaX(player.getDirection(), 1);
		int ty = py + Utils.getDeltaY(player.getDirection(), 1);

		GameTile targetTile = room.getTile(tx, ty);
		if (targetTile == null) {
			// Target tile doesn't exist
			return false;
		}

		Collection<TileAction> actions = targetTile.getActions();
		for (TileAction ta : actions) {
			if (ta.getObjectId() == objectId && ta.getId() == actionId) {
				Event event = new Event();
				event.useActionEvent(new UseActionEvent(player.getId(), objectId, actionId));
				server.executeAndSend(room, event);
				return true;
			}
		}
		
		//Action not found
		return false;
	}

	public boolean useItem(int id, int itemId) throws InvalidSecurityToken {
		Room room = server.getRoomByUser(id);
		if (room == null) {
			return false; // player was killed while trying to use an item;
		}
		GameObject object = room.getObjectById(id);
		if (!(object instanceof Player)) {
			return false; // Referenced object is not player-controlled
		}
		Player player = (Player) object;
		if (player.getHp() <= 0) {
			return false; //Player is dead
		}
		if (player.getActivity() != null) {
			return false; //Player is performing some other action
		}

		Item item = player.getInventory().getItemById(itemId);
		if (item == null || !(item instanceof UsableItem)) {
			// Item not in inventory or not a UsableItem
			return false;
		}
		UsableItem ui = (UsableItem) item;
		if (ui.canUse()) {
			// Item can't be used
			return false;
		}

		Event event = new Event();
		event.useEvent(new UseEvent(ui.getId(), player.getId()));
		server.executeAndSend(room, event);
		return true;
	}
	
	// Getters

	// Setters

}
