package mud.server.corba;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Random;

import mud.common.Direction;
import mud.common.Utils;
import mud.common.events.ServerEventHandler;
import mud.common.idl.clients.Client;
import mud.common.idl.clients.Event;
import mud.common.idl.clients.EventPackage.AddObjectEvent;
import mud.common.idl.clients.EventPackage.ChatEvent;
import mud.common.idl.clients.EventPackage.RemoveObjectEvent;
import mud.common.idl.clients.EventPackage.RoomInitEvent;
import mud.common.idl.servers.ServerPOA;
import mud.common.idl.servers.ServerPackage.InvalidSecurityToken;
import mud.common.items.Armor;
import mud.common.items.Weapon;
import mud.common.loader.RoomLoader;
import mud.common.objects.Player;
import mud.common.room.Room;
import mud.server.EventBroadcaster;
import mud.server.MonsterController;

import org.omg.CORBA.ORB;

/**
 * Implements the server, which maintains all administration of all game rooms
 * and users. Generated session ids are always lower than 1000.
 * 
 * @author Pieter van Ede
 * @author Tamar Christina
 */
public class ServerImpl extends ServerPOA {
	
	private static final int INITIAL_ROOM = 1;
	private int idGenerator = 1000;
	
	public static Random random = new Random();

	private Properties users;	
	private Map<Integer, User> loggedIn;
	private Map<Integer, RoomInfo> rooms;
	
	private MonsterController monsterController;
	private GameLogic logic;
	private EventBroadcaster messagequeue;

	public ServerImpl(ORB orb, int maxthreads) throws InstantiationException {
		// load user base
		users = new Properties();
		FileInputStream fis = null;
		try {
			fis = new FileInputStream(new File("users.xml"));
			users.loadFromXML(fis);			
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		} finally {
			try {
				if (fis != null) fis.close();
			} catch (IOException e) { }
		}

		loggedIn = new Hashtable<Integer, User>();
		rooms = new Hashtable<Integer, RoomInfo>();
		
		loadRooms();
		
		logic = new GameLogic(this);
		messagequeue = EventBroadcaster.getInstance(this, maxthreads);
		messagequeue.start();
		
		//Start monster controller thread
		Collection<Room> rs = new ArrayList<Room>(rooms.size());
		for (RoomInfo ri : rooms.values()) {
			rs.add(ri.getRoom());
		}
		monsterController = new MonsterController();
		monsterController.start(this, logic, rs);
	}

	//Functions
	public static void main(String[] args) throws IOException {
		Properties prop = new Properties();
		prop.setProperty("pjede", "pjedeww");
		FileOutputStream fos = new FileOutputStream(new File("users.xml"));
		prop.storeToXML(fos, "the comment");
		System.out.println("users.xml generated succesfully");
	}
	
	protected void loadRooms() {
		List<Room> loaded = new ArrayList<Room>();
		
		RoomLoader loader = new RoomLoader();
		for (File roomFile : new File("rooms").listFiles()) {
			if (!roomFile.getName().endsWith(".xml")) {
				continue;
			}
			
			try {
				FileInputStream fin = new FileInputStream(roomFile);
				byte[] bytes = new byte[(int)roomFile.length()];
				fin.read(bytes);
				fin.close();
				
				Room r = loader.loadRoom(new String(bytes));
				if (r != null) loaded.add(r);
			} catch (IOException ioe) {
				ioe.printStackTrace();
			}
		}
		
		for (Room r : loaded) {
			rooms.put(r.getId(), new RoomInfo(r));
			System.out.println("Loaded room: " + r.getId());
		}
	}

	public synchronized int login(char[] uname, char[] passw, Client client) {
		String username = new String(uname).trim();
		String password = new String(passw).trim();
		
		if (!users.containsKey(username) || !password.equals(users.get(username))) {
			//Invalid login
			return 0;
		}
			
		int id = getIdByUsername(username);
		if (id != 0) {
			//Already logged in
			return id;
		}
		
		//Generate user id
		do {
			id = generateId();
		} while (isLoggedIn(id));

		ClientConnection con = new ClientConnection(id, client);
		User user = new User(id, username, con);
		loggedIn.put(id, user);
		
		Player player = user.getPlayer();
		player.getInventory().addItem(new Weapon(generateId(), "Spork", 2));
		player.getInventory().addItem(new Armor(generateId(), "Lousy Armor", 1));
		setPlayerRoom(player, INITIAL_ROOM, 1, 1, player.getDirection());

		sendMessage(0, "%s has joined the server", username);
		return id;
	}
	public synchronized void logout(int id) throws InvalidSecurityToken {
		if (!loggedIn.containsKey(id)) {
			throw new InvalidSecurityToken("Can't logout when you're not logged in");
		}
			
		User user = loggedIn.get(id);		
		setPlayerRoom(user.getPlayer(), 0);
		loggedIn.remove(id);
		
		sendMessage(0, String.format("%s has been logged out", user.getName()));
	}

	public synchronized void setPlayerRoom(Player player, int roomId) {
		setPlayerRoom(player, roomId, 0, 0, player.getDirection());
	}
	
	public synchronized void setPlayerRoom(Player player, int roomId, int tx, int ty, Direction dir) {
		Room oldRoom = player.getRoom();
		if (oldRoom != null) {
			rooms.get(oldRoom.getId()).removeUser(player.getId());
			oldRoom.removeObject(player);
			
			Event event = new Event();
			event.removeObjectEvent(new RemoveObjectEvent(player.getId()));
			messagequeue.sendToRoom(oldRoom.getId(), event);
		}
		
		RoomInfo roomInfo = rooms.get(roomId);
		
		if (roomInfo != null) {
			Room room = roomInfo.getRoom();
			roomInfo.addUser(loggedIn.get(player.getId()));
			room.addObject(player);
			player.setPos(tx, ty);
			player.setDirection(dir);
					
			for (User user : roomInfo.getUsersInRoom()) {
				Event event = new Event();
				if (user.getId() == player.getId()) {
					event.roomInitEvent(new RoomInitEvent(room.toXml().toXmlString()));
				} else {
					event.addObjectEvent(new AddObjectEvent(player.toXml().toXmlString()));
				}
				messagequeue.sendToSingle(event, user.getConnection());
			}
		}
	}
	
	public void sendMessage(int id, char message[]) {
		sendMessage(id, new String(message).trim());
	}
	public void sendMessage(int id, String format, Object... args) {
		String message = String.format(format, args);
				
		String sender = null;
		int color = 0;
		if (id != 0) {
			sender = loggedIn.get(id).getName();
			color = Math.abs(sender.hashCode()) % 16777215;		
		}
		for (User user : loggedIn.values()) {
			String m = message;
			if (sender != null) {
				m = String.format("<font color='%d'>&lt;%s&gt;</font>&nbsp;%s", color, sender, message);
			}

			Event event = new Event();
			event.chatEvent(new ChatEvent(Utils.createFixedLengthCharArray(m, 256)));
			messagequeue.sendToSingle(event, user.getConnection());
		}
	}

	public void executeAndSend(Room room, Event event) {
		try {
			ServerEventHandler.process(room, event);
			messagequeue.sendToRoom(room.getId(), event);
		} catch (IOException e) {
			e.printStackTrace();
		}
	}
	
	protected int generateId() {
		idGenerator++;
		return idGenerator;
	}

	public synchronized boolean move(int id, int x, int y) throws InvalidSecurityToken {
		return logic.move(id, x, y);
	}
	public synchronized boolean action(int id, int object_id, int action_index) throws InvalidSecurityToken {
		return logic.action(id, object_id, action_index);
	}
	public synchronized boolean attack(int id) throws InvalidSecurityToken {
		return logic.attack(id);
	}
	public synchronized boolean useItem(int id, int item_id) throws InvalidSecurityToken {
		return logic.useItem(id, item_id);
	}

	//Getters
	public boolean isLoggedIn(int id) {
		return loggedIn.containsKey(id);
	}
	private int getIdByUsername(String name) {
		for (User u : loggedIn.values()) {
			if (name.equals(u.getName())) {
				return u.getId();
			}
		}
		return 0;
	}	
	public User getUser(int userId) {
		return loggedIn.get(userId);
	}
	public Room getRoomByUser(int userId) {
		return loggedIn.get(userId).getPlayer().getRoom();
	}
	public synchronized ClientConnection[] getAllConnections() {
		ClientConnection out[] = new ClientConnection[users.size()];
		int t = 0;
		for (User user : loggedIn.values()) {
			out[t] = user.getConnection();
			t++;
		}
		return out;
	}
	public synchronized ClientConnection[] getConnectionsByRoom(int roomId) {
		RoomInfo ri = rooms.get(roomId);
		if (ri != null) {
			Collection<User> users = ri.getUsersInRoom();
			ClientConnection out[] = new ClientConnection[users.size()];
			int t = 0;
			for (User user : users) {
				out[t] = user.getConnection();
				t++;
			}
			return out;
		}
		return new ClientConnection[0];
	}

	public synchronized char[][] getUsers(int id) throws InvalidSecurityToken {
		char usernames[][] = new char[loggedIn.size()][];
		
		int t = 0;
		for (User user : loggedIn.values()) {
			usernames[t] = Utils.createFixedLengthCharArray(user.getName(),64);
			t++;
		}
		
		return usernames;
	}

	//Setters
}
